class Pygtail:
    def __init__(self, filename: str, offset_file: str = None, paranoid: bool = False, copytruncate: bool = True,
                 every_n: int = 0, on_update: bool = False, read_from_end: bool = False, log_patterns: list = None):
        ...
    def __iter__(self):
        ...

def __iter__():
    ...
